__author__ = 'nico'

import logging
import asyncio
from urllib.parse import urlparse

from transitions import Machine, MachineError

from hbmqtt.utils import not_in_dict_or_none
from hbmqtt.session import Session, SessionState
from hbmqtt.mqtt.connack import ReturnCode
from hbmqtt.mqtt.subscribe import SubscribePacket
from hbmqtt.mqtt.suback import SubackPacket
from hbmqtt.errors import MQTTException
from hbmqtt.mqtt.protocol import ClientProtocolHandler

_defaults = {
    'keep_alive': 10,
    'ping_delay': 1,
    'default_qos': 0,
    'default_retain': False,
    'inflight-polling-interval': 1,
    'subscriptions-polling-interval': 1,
}


class ClientException(BaseException):
    pass


def gen_client_id():
    import uuid
    return str(uuid.uuid4())


class MQTTClient:
    states = ['new', 'connecting', 'connected', 'disconnected']

    def __init__(self, client_id=None, config={}, loop=None):
        """

        :param config: Example yaml config
            broker:
                host: localhost
                port: 1883
                scheme: mqtt
                username: xxx
                password: yyy
                # OR
                uri: mqtt:xxx@yyy//localhost:1883/
                # OR a mix ot both
            keep_alive: 60
            cleansession: true
            will:
                retain: false
                topic: some/topic
                message: Will message
                qos: 0
            default_qos: 0
            default_retain: false
            topics:
                a/b:
                    qos: 2
                    retain: true
        :param loop:
        :return:
        """
        self.logger = logging.getLogger(__name__)
        self.config = config.copy()
        self.config.update(_defaults)
        if client_id is not None:
            self.client_id = client_id
        else:
            self.client_id = gen_client_id()
            self.logger.debug("Using generated client ID : %s" % self.client_id)

        self._init_states()
        if loop is not None:
            self._loop = loop
        else:
            self._loop = asyncio.get_event_loop()
        self.session = None
        self._handler = None

    def _init_states(self):
        self.machine = Machine(states=MQTTClient.states, initial='new')
        self.machine.add_transition(trigger='connect', source='new', dest='connecting')
        self.machine.add_transition(trigger='connect', source='disconnected', dest='connecting')
        self.machine.add_transition(trigger='connect_fail', source='connecting', dest='disconnected')
        self.machine.add_transition(trigger='connect_success', source='connecting', dest='connected')
        self.machine.add_transition(trigger='disconnect', source='idle', dest='disconnected')
        self.machine.add_transition(trigger='disconnect', source='connected', dest='disconnected')

    @asyncio.coroutine
    def connect(self, host=None, port=None, username=None, password=None, uri=None, cleansession=None):
        try:
            self.machine.connect()
            self.session = self._initsession(host, port, username, password, uri, cleansession)
            self.logger.debug("Connect with session parameters: %s" % self.session)

            yield from self._connect_coro()
            self.machine.connect_success()
        except MachineError:
            msg = "Connect call incompatible with client current state '%s'" % self.machine.current_state
            self.logger.warn(msg)
            self.machine.connect_fail()
            raise ClientException(msg)
        except Exception as e:
            self.machine.connect_fail()
            self.logger.warn("Connection failed: %s " % e)
            raise ClientException("Connection failed: %s " % e)

    @asyncio.coroutine
    def disconnect(self):
        try:
            yield from self._handler.mqtt_disconnect()
            yield from self._handler.stop()
        except Exception as e:
            self.logger.warn("Unhandled exception: %s" % e)
            raise ClientException("Unhandled exception: %s" % e)
        except MachineError as me:
            self.logger.debug("Invalid method call at this moment: %s" % me)
            raise ClientException("Client instance can't be disconnected: %s" % me)
        self.session = None

    @asyncio.coroutine
    def ping(self):
        """
        Send a MQTT ping request and wait for response
        :return: None
        """
        self._handler.mqtt_ping()

    @asyncio.coroutine
    def publish(self, topic, message, dup=False, qos=None, retain=None):
        def get_retain_and_qos():
            if qos:
                _qos = qos
            else:
                _qos = self.config['default_qos']
                try:
                    _qos = self.config['topics'][topic]['qos']
                except KeyError:
                    pass
            if retain:
                _retain = retain
            else:
                _retain = self.config['default_retain']
                try:
                    _retain = self.config['topics'][topic]['retain']
                except KeyError:
                    pass
            return _qos, _retain
        (app_qos, app_retain) = get_retain_and_qos()
        if app_qos == 0:
            yield from self._publish_qos_0(topic, message, dup, app_retain)
        if app_qos == 1:
            yield from self._publish_qos_1(topic, message, dup, app_retain)
        if app_qos == 2:
            yield from self._publish_qos_2(topic, message, dup, app_retain)

    @asyncio.coroutine
    def _publish_qos_0(self, topic, message, dup, retain):
        yield from self._handler.mqtt_publish(topic, message, self.session.next_packet_id, dup, 0x00, retain)

    @asyncio.coroutine
    def _publish_qos_1(self, topic, message, dup, retain):
        yield from self._handler.mqtt_publish(topic, message, self.session.next_packet_id, dup, 0x01, retain)

    @asyncio.coroutine
    def _publish_qos_2(self, topic, message, dup, retain):
        yield from self._handler.mqtt_publish(topic, message, self.session.next_packet_id, dup, 0x02, retain)

    @asyncio.coroutine
    def subscribe(self, topics):
        yield from self._handler.mqtt_subscribe(topics, self.session.next_packet_id)

    @asyncio.coroutine
    def unsubscribe(self, topics):
        yield from self._handler.mqtt_unsubscribe(topics, self.session.next_packet_id)

    @asyncio.coroutine
    def _connect_coro(self):
        try:
            self.session.reader, self.session.writer = \
                yield from asyncio.open_connection(self.session.remote_address, self.session.remote_port)
            self._handler = ClientProtocolHandler(self.session, self.config)
            yield from self._handler.start()

            return_code = yield from self._handler.mqtt_connect()

            if return_code is not ReturnCode.CONNECTION_ACCEPTED:
                raise ClientException("Connection rejected with code '%s'" % hex(return_code))

            self.session.state = SessionState.CONNECTED
            self.logger.debug("connected to %s:%s" % (self.session.remote_address, self.session.remote_port))
        except Exception as e:
            self.session.state = SessionState.DISCONNECTED
            raise e

    def _initsession(self, host=None, port=None, username=None, password=None, uri=None, cleansession=None) -> dict:
        # Load config
        broker_conf = self.config.get('broker', dict()).copy()
        if 'mqtt' not in broker_conf:
            broker_conf['scheme'] = 'mqtt'
        if 'username' not in broker_conf:
            broker_conf['username'] = None
        if 'password' not in broker_conf:
            broker_conf['password'] = None

        if uri is not None:
            result = urlparse(uri)
            if result.scheme:
                broker_conf['scheme'] = result.scheme
            if result.hostname:
                broker_conf['host'] = result.hostname
            if result.port:
                broker_conf['port'] = result.port
            if result.username:
                broker_conf['username'] = result.username
            if result.password:
                broker_conf['password'] = result.password
        if host:
            broker_conf['host'] = host
        if port:
            broker_conf['port'] = int(port)
        if username:
            broker_conf['username'] = username
        if password:
            broker_conf['password'] = password
        if cleansession is not None:
            broker_conf['cleansession'] = cleansession

        for key in ['scheme', 'host', 'port']:
            if not_in_dict_or_none(broker_conf, key):
                raise ClientException("Missing connection parameter '%s'" % key)

        s = Session()
        s.client_id = self.client_id
        s.remote_address = broker_conf['host']
        s.remote_port = broker_conf['port']
        s.username = broker_conf['username']
        s.password = broker_conf['password']
        s.scheme = broker_conf['scheme']
        if cleansession is not None:
            s.cleansession = cleansession
        else:
            s.cleansession = self.config.get('cleansession', True)
        s.keep_alive = self.config['keep_alive']
        if 'will' in self.config:
            s.will_flag = True
            s.will_retain = self.config['will']['retain']
            s.will_topic = self.config['will']['topic']
            s.will_message = self.config['will']['message']
        else:
            s.will_flag = False
            s.will_retain = False
            s.will_topic = None
            s.will_message = None
        return s




# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
from hbmqtt.mqtt.packet import MQTTPacket, MQTTFixedHeader, PacketType, MQTTVariableHeader, MQTTPayload
from hbmqtt.codecs import *
from hbmqtt.errors import MQTTException, CodecException, HBMQTTException, NoDataException


class ConnectVariableHeader(MQTTVariableHeader):
    USERNAME_FLAG = 0x80
    PASSWORD_FLAG = 0x40
    WILL_RETAIN_FLAG = 0x20
    WILL_FLAG = 0x04
    WILL_QOS_MASK = 0x18
    CLEAN_SESSION_FLAG = 0x02
    RESERVED_FLAG = 0x01

    def __init__(self, connect_flags=0x00, keep_alive=0, proto_name='MQTT', proto_level=0x04):
        super().__init__()
        self.proto_name = proto_name
        self.proto_level = proto_level
        self.flags = connect_flags
        self.keep_alive = keep_alive

    def __repr__(self):
        return "ConnectVariableHeader(proto_name={0}, proto_level={1}, flags={2}, keepalive={3})".format(
            self.proto_name, self.proto_level, hex(self.flags), self.keep_alive)

    def _set_flag(self, val, mask):
        if val:
            self.flags |= mask
        else:
            self.flags &= ~mask

    def _get_flag(self, mask):
        if self.flags & mask:
            return True
        else:
            return False

    @property
    def username_flag(self) -> bool:
        return self._get_flag(self.USERNAME_FLAG)

    @username_flag.setter
    def username_flag(self, val: bool):
        self._set_flag(val, self.USERNAME_FLAG)

    @property
    def password_flag(self) -> bool:
        return self._get_flag(self.PASSWORD_FLAG)

    @password_flag.setter
    def password_flag(self, val: bool):
        self._set_flag(val, self.PASSWORD_FLAG)

    @property
    def will_retain_flag(self) -> bool:
        return self._get_flag(self.WILL_RETAIN_FLAG)

    @will_retain_flag.setter
    def will_retain_flag(self, val: bool):
        self._set_flag(val, self.WILL_RETAIN_FLAG)

    @property
    def will_flag(self) -> bool:
        return self._get_flag(self.WILL_FLAG)

    @will_flag.setter
    def will_flag(self, val: bool):
        self._set_flag(val, self.WILL_FLAG)

    @property
    def clean_session_flag(self) -> bool:
        return self._get_flag(self.CLEAN_SESSION_FLAG)

    @clean_session_flag.setter
    def clean_session_flag(self, val: bool):
        self._set_flag(val, self.CLEAN_SESSION_FLAG)

    @property
    def reserved_flag(self) -> bool:
        return self._get_flag(self.RESERVED_FLAG)

    @property
    def will_qos(self):
        return (self.flags & 0x18) >> 3

    @will_qos.setter
    def will_qos(self, val: int):
        self.flags &= (0x00 << 3)
        self.flags |= (val << 3)

    @classmethod
    @asyncio.coroutine
    def from_stream(cls, reader: asyncio.StreamReader, fixed_header: MQTTFixedHeader):
        #  protocol name
        protocol_name = yield from decode_string(reader)
        if protocol_name != "MQTT":
            raise MQTTException('[MQTT-3.1.2-1] Incorrect protocol name: "%s"' % protocol_name)

        # protocol level (only MQTT 3.1.1 supported)
        protocol_level_byte = yield from read_or_raise(reader, 1)
        protocol_level = bytes_to_int(protocol_level_byte)

        # flags
        flags_byte = yield from read_or_raise(reader, 1)
        flags = bytes_to_int(flags_byte)
        if flags & 0x01:
            raise MQTTException('[MQTT-3.1.2-3] CONNECT reserved flag must be set to 0')

        # keep-alive
        keep_alive_byte = yield from read_or_raise(reader, 2)
        keep_alive = bytes_to_int(keep_alive_byte)

        return cls(flags, keep_alive, protocol_name, protocol_level)

    def to_bytes(self):
        out = b''

        # Protocol name
        out += encode_string(self.proto_name)
        # Protocol level
        out += int_to_bytes(self.proto_level)
        # flags
        out += int_to_bytes(self.flags)
        # keep alive
        out += int_to_bytes(self.keep_alive, 2)

        return out


class ConnectPayload(MQTTPayload):
    def __init__(self, client_id=None, will_topic=None, will_message=None, username=None, password=None):
        super().__init__()
        self.client_id = client_id
        self.will_topic = will_topic
        self.will_message = will_message
        self.username = username
        self.password = password

    def __repr__(self):
        return "ConnectVariableHeader(client_id={0}, will_topic={1}, will_message={2}, username={3}, password={4})".format(
            self.client_id, self.will_topic, self.will_message, self.username, self.password)

    @classmethod
    @asyncio.coroutine
    def from_stream(cls, reader: asyncio.StreamReader, fixed_header: MQTTFixedHeader,
                    variable_header: ConnectVariableHeader):
        payload = cls()
        #  Client identifier
        try:
            payload.client_id = yield from decode_string(reader)
        except NoDataException:
            raise MQTTException('[[MQTT-3.1.3-3]] Client identifier must be present')

        # Read will topic, username and password
        if variable_header.will_flag:
            try:
                payload.will_topic = yield from decode_string(reader)
                payload.will_message = yield from decode_string(reader)
            except NoDataException:
                raise MQTTException('will flag set, but will topic/message not present in payload')

        if variable_header.username_flag:
            try:
                payload.username = yield from decode_string(reader)
            except NoDataException:
                raise MQTTException('username flag set, but username not present in payload')

        if variable_header.password_flag:
            try:
                payload.password = yield from decode_string(reader)
            except NoDataException:
                raise MQTTException('password flag set, but password not present in payload')

        return payload

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: ConnectVariableHeader):
        out = b''
        # Client identifier
        out += encode_string(self.client_id)
        # Will topic / message
        if variable_header.will_flag:
            out += encode_string(self.will_topic)
            out += encode_string(self.will_message)
        # username
        if variable_header.username_flag:
            out += encode_string(self.username)
        # password
        if variable_header.password_flag:
            out += encode_string(self.password)

        return out


class ConnectPacket(MQTTPacket):
    VARIABLE_HEADER = ConnectVariableHeader
    PAYLOAD = ConnectPayload

    def __init__(self, fixed: MQTTFixedHeader=None, vh: ConnectVariableHeader=None, payload: ConnectPayload=None):
        if fixed is None:
            header = MQTTFixedHeader(PacketType.CONNECT, 0x00)
        else:
            if fixed.packet_type is not PacketType.CONNECT:
                raise HBMQTTException("Invalid fixed packet type %s for ConnectPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = vh
        self.payload = payload



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest
import asyncio

from hbmqtt.mqtt.connect import ConnectPacket, ConnectVariableHeader, ConnectPayload
from hbmqtt.mqtt.protocol import ProtocolHandler
from hbmqtt.session import Session
from hbmqtt.mqtt.packet import PacketType
import logging

logging.basicConfig(level=logging.DEBUG)

ret_packet = None

config = {
    'keep_alive': 10,
    'ping_delay': 1,
    'default_qos': 0,
    'default_retain': False,
    'inflight-polling-interval': 1,
    'subscriptions-polling-interval': 1,
}

class ConnectPacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        self.logger = logging.getLogger(__name__)

    def test_read_loop(self):
        data = b'\x10\x3e\x00\x04MQTT\x04\xce\x00\x00\x00\x0a0123456789\x00\x09WillTopic\x00\x0bWillMessage\x00\x04user\x00\x08password'
        @asyncio.coroutine
        def serve_test(reader, writer):
            writer.write(data)
            yield from writer.drain()
            writer.close()

        loop = asyncio.get_event_loop()
        coro = asyncio.start_server(serve_test, '127.0.0.1', 8888, loop=loop)
        server = loop.run_until_complete(coro)

        @asyncio.coroutine
        def client():
            S = Session()
            S.reader, S.writer = yield from asyncio.open_connection('127.0.0.1', 8888,
                                                        loop=loop)
            h = ProtocolHandler(S, config)
            yield from h.start()
            incoming_packet = yield from h.incoming_queues[PacketType.CONNECT].get()
            yield from h.stop()
            return incoming_packet

        packet = loop.run_until_complete(client())
        server.close()
        self.assertEquals(packet.fixed_header.packet_type, PacketType.CONNECT)

    def test_write_loop(self):
        test_packet = ConnectPacket(vh=ConnectVariableHeader(), payload=ConnectPayload('Id', 'WillTopic', 'WillMessage', 'user', 'password'))
        event=asyncio.Event()

        @asyncio.coroutine
        def serve_test(reader, writer):
            global ret_packet
            packet = yield from ConnectPacket.from_stream(reader)
            ret_packet = packet
            writer.close()
            event.set()

        @asyncio.coroutine
        def client():
            S = Session()
            S.reader, S.writer = yield from asyncio.open_connection('127.0.0.1', 8888, loop=loop)
            h = ProtocolHandler(S, config)
            yield from h.start()
            yield from h.outgoing_queue.put(test_packet)
            yield from h.stop()

        # Start server
        loop = asyncio.get_event_loop()
        coro = asyncio.start_server(serve_test, '127.0.0.1', 8888, loop=loop)
        server = loop.run_until_complete(coro)

        # Schedule client
        loop.call_soon(asyncio.async, client())

        # Wait for server to complete client request
        loop.run_until_complete(asyncio.wait([event.wait()]))
        server.close()
        self.logger.info(ret_packet)
        self.assertEquals(ret_packet.fixed_header.packet_type, PacketType.CONNECT)


# # Copyright (c) 2015 Nicolas JOUANIN
# #
# # See the file license.txt for copying permission.
#
# import asyncio
#
# from hbmqtt.messages.packet import MQTTMessage, ConnectMessage, ConnackMessage
# from hbmqtt.session import Session, ClientState
# from hbmqtt.errors import BrokerException,MQTTException
# from hbmqtt.broker.handlers.abstract import AbstractHandler
#
#
# class ConnectException(CodecException):
#     pass
#
# class ConnectHandler(AbstractHandler):
#     def __init__(self, broker):
#         super().__init__(broker)
#
#     def _check_header(self, header):
#         if header.flags:
#             raise MQTTException("[MQTT-2.2.2-1] Header flags reserved for future use")
#
#     def _decode_payload(self, session: Session, header: MQTTHeader):
#
#     @asyncio.coroutine
#     def handle(self, session: Session, message: ConnectMessage) -> MQTTMessage:
#         response = None
#
#         # Check Protocol
#         # protocol level (only MQTT 3.1.1 supported)
#         if message.proto_level != 4:
#             return ConnackMessage(False, ConnackMessage.ReturnCode.UNACCEPTABLE_PROTOCOL_VERSION)
#
#         # No client ID provided
#         if message.client_id is None or message.client_id == "":
#             if message.is_clean_session():
#                 # [MQTT-3.1.3-6] and [MQTT-3.1.3-7]
#                 message.client_id = self.gen_client_id()
#             else:
#                 # [MQTT-3.1.3-8] : Identifier rejected
#                 return ConnackMessage(False, ConnackMessage.ReturnCode.IDENTIFIER_REJECTED)
#
#         if message.is_clean_session():
#             try:
#                 self.broker.discard_session(message.client_id)
#             except BrokerException:
#                 pass
#             session = self.broker.create_session(message._remote_address, message._remote_port, message.client_id, message.is_clean_session())
#             # [MQTT-3.2.2-1]
#             response = ConnackMessage(False, 0)
#         else:
#             try:
#                 session = self.broker.get_session(message.client_id)
#                 if session.client_state == ClientState.CONNECTED:
#                     # [MQTT-3.1.4-2]
#                         # TODO : Add logging
#                         return ConnackMessage(False, ConnackMessage.ReturnCode.IDENTIFIER_REJECTED)
#                 else:
#                     # [MQTT-3.2.2-2]
#                     response = ConnackMessage(True, 0)
#             except BrokerException:
#                 session = self.broker.create_session(message._remote_address, message._remote_port, message.client_id, message.is_clean_session())
#                 response = ConnackMessage(False, 0)
#             if session.client_state == ClientState.DISCONNECTED:
#                 self.broker.resume_session(session)
#                 session.client_state = ClientState.CONNECTED
#
#         if message.is_will_flag():
#             session.will_flag = True
#             session.will_message = message.will_message
#             session.will_qos = message.will_qos()
#             session.will_retain = message.is_will_retain()
#
#         session.keep_alive = message.keep_alive
#
#         self.broker.save_session(session)
#         return response
#
#     def gen_client_id(self):
#         import uuid
#         return uuid.uuid4()


__author__ = 'nico'

import abc
import asyncio

from hbmqtt.messages.packet import MQTTHeader
from hbmqtt.session import Session


class AbstractHandler(metaclass=abc.ABCMeta):
    def __init__(self, broker):
        self._broker = broker

    @abc.abstractmethod
    @asyncio.coroutine
    def _check_header(self, header):
        return

    @abc.abstractmethod
    @asyncio.coroutine
    def _decode_payload(self, session: Session, header: MQTTHeader):
        return

    @abc.abstractmethod
    @asyncio.coroutine
    def _process_request(self, session: Session, message):
        return

    @asyncio.coroutine
    def process_incoming_packet(self, session: Session, header: MQTTHeader):
        yield from self._check_header(header)
        message = yield from self._decode_payload(session, header)
        yield from self._process_request(session, message)


# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import logging
import asyncio
from hbmqtt.mqtt.packet import MQTTFixedHeader
from hbmqtt.mqtt import packet_class
from hbmqtt.errors import NoDataException
from hbmqtt.mqtt.packet import PacketType
from hbmqtt.mqtt.connect import ConnectVariableHeader, ConnectPacket, ConnectPayload
from hbmqtt.mqtt.disconnect import DisconnectPacket
from hbmqtt.mqtt.pingreq import PingReqPacket
from hbmqtt.mqtt.publish import PublishPacket
from hbmqtt.mqtt.pubrel import PubrelPacket
from hbmqtt.mqtt.subscribe import SubscribePacket
from hbmqtt.mqtt.unsubscribe import UnsubscribePacket
from hbmqtt.session import Session
from blinker import Signal
from transitions import Machine, MachineError

class InFlightMessage:
    states = ['new', 'published', 'acknowledged', 'received', 'released', 'completed']

    def __init__(self, packet_id, qos):
        self.packet_id = packet_id
        self.qos = qos
        self._init_states()

    def _init_states(self):
        self.machine = Machine(model=self, states=InFlightMessage.states, initial='new')
        self.machine.add_transition(trigger='publish', source='new', dest='published')
        if self.qos == 0x01:
            self.machine.add_transition(trigger='acknowledge', source='published', dest='acknowledged')
        if self.qos == 0x02:
            self.machine.add_transition(trigger='receive', source='published', dest='received')
            self.machine.add_transition(trigger='release', source='received', dest='released')
            self.machine.add_transition(trigger='complete', source='released', dest='completed')


class ProtocolHandler:
    """
    Class implementing the MQTT communication protocol using asyncio features
    """
    packet_sent = Signal()
    packet_received = Signal()

    def __init__(self, session: Session, config, loop=None):
        self.logger = logging.getLogger(__name__)
        self.session = session
        self.config = config
        if loop is None:
            self._loop = asyncio.get_event_loop()
        else:
            self._loop = loop
        self._reader_task = None
        self._writer_task = None
        self._inflight_task = None
        self._reader_ready = asyncio.Event(loop=self._loop)
        self._writer_ready = asyncio.Event(loop=self._loop)
        self._inflight_ready = asyncio.Event(loop=self._loop)
        self._inflight_changed = asyncio.Condition(loop=self._loop)

        self._running = False

        self.session.local_address, self.session.local_port = self.session.writer.get_extra_info('sockname')

        self.incoming_queues = dict()
        for p in PacketType:
            self.incoming_queues[p] = asyncio.Queue()
        self.outgoing_queue = asyncio.Queue()
        self.inflight_messages = dict()

    @asyncio.coroutine
    def start(self):
        self._running = True
        self._reader_task = asyncio.async(self._reader_coro(), loop=self._loop)
        self._writer_task = asyncio.async(self._writer_coro(), loop=self._loop)
        self._inflight_task = asyncio.async(self._inflight_coro(), loop=self._loop)
        yield from asyncio.wait(
            [self._reader_ready.wait(), self._writer_ready.wait(), self._inflight_ready.wait()], loop=self._loop)
        self.logger.debug("Handler tasks started")

    @asyncio.coroutine
    def mqtt_publish(self, topic, message, packet_id, dup, qos, retain):
        def qos_0_predicate():
            ret = False
            try:
                if self.inflight_messages.get(packet_id).state == 'published':
                    ret = True
                #self.logger.debug("qos_0 predicate return %s" % ret)
                return ret
            except KeyError:
                return False

        def qos_1_predicate():
            ret = False
            try:
                if self.inflight_messages.get(packet_id).state == 'acknowledged':
                    ret = True
                #self.logger.debug("qos_1 predicate return %s" % ret)
                return ret
            except KeyError:
                return False

        def qos_2_predicate():
            ret = False
            try:
                if self.inflight_messages.get(packet_id).state == 'completed':
                    ret = True
                #self.logger.debug("qos_1 predicate return %s" % ret)
                return ret
            except KeyError:
                return False

        if packet_id in self.inflight_messages:
            self.logger.warn("A message with the same packet ID is already in flight")
        packet = PublishPacket.build(topic, message, packet_id, dup, qos, retain)
        yield from self.outgoing_queue.put(packet)
        inflight_message = InFlightMessage(packet.variable_header.packet_id, qos)
        inflight_message.publish()
        self.inflight_messages[packet.variable_header.packet_id] = inflight_message
        yield from self._inflight_changed.acquire()
        if qos == 0x00:
            yield from self._inflight_changed.wait_for(qos_0_predicate)
        if qos == 0x01:
            yield from self._inflight_changed.wait_for(qos_1_predicate)
        if qos == 0x02:
            yield from self._inflight_changed.wait_for(qos_2_predicate)
        self.inflight_messages.pop(packet.variable_header.packet_id)
        self._inflight_changed.release()
        return packet

    @asyncio.coroutine
    def stop(self):
        self._running = False
        self.session.reader.feed_eof()
        yield from asyncio.wait([self._inflight_task, self._writer_task, self._reader_task], loop=self._loop)

    @asyncio.coroutine
    def _reader_coro(self):
        self.logger.debug("Starting reader coro")
        while self._running:
            try:
                self._reader_ready.set()
                fixed_header = yield from asyncio.wait_for(MQTTFixedHeader.from_stream(self.session.reader), 5)
                if fixed_header:
                    cls = packet_class(fixed_header)
                    packet = yield from cls.from_stream(self.session.reader, fixed_header=fixed_header)
                    self.logger.debug(" <-in-- " + repr(packet))
                    yield from self.incoming_queues[packet.fixed_header.packet_type].put(packet)
                    self.packet_received.send(packet)
                else:
                    self.logger.debug("No more data, stopping reader coro")
                    break
            except asyncio.TimeoutError:
                self.logger.debug("Input stream read timeout")
            except NoDataException as nde:
                self.logger.debug("No data available")
            except Exception as e:
                self.logger.warn("Unhandled exception in reader coro: %s" % e)
                break
        self.logger.debug("Reader coro stopped")

    @asyncio.coroutine
    def _writer_coro(self):
        self.logger.debug("Starting writer coro")
        while self._running:
            try:
                self._writer_ready.set()
                packet = yield from asyncio.wait_for(self.outgoing_queue.get(), 5)
                yield from packet.to_stream(self.session.writer)
                self.logger.debug(" -out-> " + repr(packet))
                yield from self.session.writer.drain()
                self.packet_sent.send(packet)
            except asyncio.TimeoutError as ce:
                self.logger.debug("Output queue get timeout")
            except Exception as e:
                self.logger.warn("Unhandled exception in writer coro: %s" % e)
                break
        self.logger.debug("Writer coro stopping")
        # Flush queue before stopping
        if not self.outgoing_queue.empty():
            while True:
                try:
                    packet = self.outgoing_queue.get_nowait()
                    yield from packet.to_stream(self.session.writer)
                    self.logger.debug(" -out-> " + repr(packet))
                except asyncio.QueueEmpty:
                    break
                except Exception as e:
                    self.logger.warn("Unhandled exception in writer coro: %s" % e)
        self.logger.debug("Writer coro stopped")

    @asyncio.coroutine
    def _inflight_coro(self):
        self.logger.debug("Starting in-flight messages polling coro")
        while self._running:
            self._inflight_ready.set()
            yield from asyncio.sleep(self.config['inflight-polling-interval'])
            self.logger.debug("in-flight polling coro wake-up")
            try:
                while not self.incoming_queues[PacketType.PUBACK].empty():
                    packet = self.incoming_queues[PacketType.PUBACK].get_nowait()
                    packet_id = packet.variable_header.packet_id
                    inflight_message = self.inflight_messages.get(packet_id)
                    inflight_message.acknowledge()
                    self.logger.debug("Message with packet Id=%s acknowledged" % packet_id)

                while not self.incoming_queues[PacketType.PUBREC].empty():
                    packet = self.incoming_queues[PacketType.PUBREC].get_nowait()
                    packet_id = packet.variable_header.packet_id
                    inflight_message = self.inflight_messages.get(packet_id)
                    inflight_message.receive()
                    self.logger.debug("Message with packet Id=%s received" % packet_id)

                    rel_packet = PubrelPacket.build(packet_id)
                    yield from self.outgoing_queue.put(rel_packet)
                    inflight_message.release()
                    self.logger.debug("Message with packet Id=%s released" % packet_id)

                while not self.incoming_queues[PacketType.PUBCOMP].empty():
                    packet = self.incoming_queues[PacketType.PUBCOMP].get_nowait()
                    packet_id = packet.variable_header.packet_id
                    inflight_message = self.inflight_messages.get(packet_id)
                    inflight_message.complete()
                    self.logger.debug("Message with packet Id=%s completed" % packet_id)

                yield from self._inflight_changed.acquire()
                self._inflight_changed.notify_all()
                self._inflight_changed.release()
            except KeyError:
                self.logger.warn("Received %s for unknown inflight message Id %d" % (packet.fixed_header.packet_type, packet_id))
            except MachineError as me:
                self.logger.warn("Packet type incompatible with message QOS: %s" % me)
        self.logger.debug("In-flight messages polling coro stopped")


class Subscription:
    states = ['new', 'subscribed', 'acknowledged']

    def __init__(self, packet_id, topics):
        self.topics = topics
        self.packet_id = packet_id
        self._init_states()

    def _init_states(self):
        self.machine = Machine(model=self, states=Subscription.states, initial='new')
        self.machine.add_transition(trigger='subscribe', source='new', dest='subscribed')
        self.machine.add_transition(trigger='acknowledge', source='subscribed', dest='acknowledged')

class UnSubscription:
    states = ['new', 'unsubscribed', 'acknowledged']

    def __init__(self, packet_id, topics):
        self.topics = topics
        self.packet_id = packet_id
        self._init_states()

    def _init_states(self):
        self.machine = Machine(model=self, states=UnSubscription.states, initial='new')
        self.machine.add_transition(trigger='unsubscribe', source='new', dest='unsubscribed')
        self.machine.add_transition(trigger='acknowledge', source='unsubscribed', dest='acknowledged')


class ClientProtocolHandler(ProtocolHandler):
    def __init__(self, session: Session, config, loop=None):
        super().__init__(session, config, loop)
        self._ping_task = None
        self.subscriptions = dict()
        self._subscription_task = None
        self._subscriptions_changed = asyncio.Condition(loop=self._loop)
        self._subscriptions_ready = asyncio.Event(loop=self._loop)

    @asyncio.coroutine
    def start(self):
        yield from super().start()
        self.packet_sent.connect(self._do_keepalive)
        self._subscription_task = asyncio.async(self._subscriptions_coro(), loop=self._loop)
        yield from asyncio.wait([self._subscriptions_ready.wait()], loop=self._loop)

    @asyncio.coroutine
    def stop(self):
        yield from super().stop()
        yield from asyncio.wait([self._subscription_task], loop=self._loop)
        if self._ping_task:
            try:
                self._ping_task.cancel()
            except Exception:
                pass

    def _do_keepalive(self, message):
        if self._ping_task:
            try:
                self._ping_task.cancel()
            except Exception:
                pass
        next_ping = self.session.keep_alive - self.config['ping_delay']
        if next_ping > 0:
            self.logger.debug('Next ping in %d seconds if no new messages between' % next_ping)
            self._ping_task = self._loop.call_later(next_ping, asyncio.async, self.mqtt_ping())

    def _subscriptions_coro(self):
        self.logger.debug("Starting subscriptions polling coro")
        while self._running:
            self._subscriptions_ready.set()
            yield from asyncio.sleep(self.config['subscriptions-polling-interval'])
            self.logger.debug("Subscriptions polling coro wake-up")
            try:
                while not self.incoming_queues[PacketType.SUBACK].empty():
                    packet = self.incoming_queues[PacketType.SUBACK].get_nowait()
                    packet_id = packet.variable_header.packet_id
                    subscription = self.subscriptions.get(packet_id)
                    for i in range(len(subscription.topics)):
                        subscription.topics[i]['return_code'] = packet.payload.return_codes[i]
                    subscription.acknowledge()
                    self.logger.debug("Subscription with packet Id=%s acknowledged" % packet_id)

                while not self.incoming_queues[PacketType.UNSUBACK].empty():
                    packet = self.incoming_queues[PacketType.UNSUBACK].get_nowait()
                    packet_id = packet.variable_header.packet_id
                    subscription = self.subscriptions.get(packet_id)
                    subscription.acknowledge()
                    self.logger.debug("Unsubscription with packet Id=%s acknowledged" % packet_id)

                yield from self._subscriptions_changed.acquire()
                self._subscriptions_changed.notify_all()
                self._subscriptions_changed.release()
            except KeyError:
                self.logger.warn("Received %s for unknown subscription message Id %d" % (packet.fixed_header.packet_type, packet_id))
            except MachineError as me:
                self.logger.warn("Packet type incompatible with message QOS: %s" % me)
        self.logger.debug("Subscriptions polling coro stopped")

    @asyncio.coroutine
    def mqtt_subscribe(self, topics, packet_id):
        """

        :param topics: array of topics [{'filter':'/a/b', 'qos': 0x00}, ...]
        :return:
        """
        def acknowledged_predicate():
            if self.subscriptions[subscribe.variable_header.packet_id].state == 'acknowledged':
                return True
            else:
                return False

        subscribe = SubscribePacket.build(topics, packet_id)
        yield from self.outgoing_queue.put(subscribe)
        subscription = Subscription(subscribe.variable_header.packet_id, topics)
        subscription.subscribe()
        self.subscriptions[subscribe.variable_header.packet_id] = subscription
        yield from self._subscriptions_changed.acquire()
        yield from self._subscriptions_changed.wait_for(acknowledged_predicate)
        subscription = self.subscriptions.pop(subscribe.variable_header.packet_id)
        self._subscriptions_changed.release()
        return subscription

    @asyncio.coroutine
    def mqtt_unsubscribe(self, topics, packet_id):
        """

        :param topics: array of topics ['/a/b', ...]
        :return:
        """
        def acknowledged_predicate():
            if self.subscriptions[unsubscribe.variable_header.packet_id].state == 'acknowledged':
                return True
            else:
                return False

        unsubscribe = UnsubscribePacket.build(topics, packet_id)
        yield from self.outgoing_queue.put(unsubscribe)
        subscription = UnSubscription(unsubscribe.variable_header.packet_id, topics)
        subscription.unsubscribe()
        self.subscriptions[unsubscribe.variable_header.packet_id] = subscription
        self.subscriptions[unsubscribe.variable_header.packet_id] = subscription
        yield from self._subscriptions_changed.acquire()
        yield from self._subscriptions_changed.wait_for(acknowledged_predicate)
        subscription = self.subscriptions.pop(unsubscribe.variable_header.packet_id)
        self._subscriptions_changed.release()
        return subscription

    @asyncio.coroutine
    def mqtt_connect(self):
        def build_connect_packet(session):
            vh = ConnectVariableHeader()
            payload = ConnectPayload()

            vh.keep_alive = session.keep_alive
            vh.clean_session_flag = session.clean_session
            vh.will_retain_flag = session.will_retain
            payload.client_id = session.client_id

            if session.username:
                vh.username_flag = True
                payload.username = session.username
            else:
                vh.username_flag = False

            if session.password:
                vh.password_flag = True
                payload.password = session.password
            else:
                vh.password_flag = False
            if session.will_flag:
                vh.will_flag = True
                vh.will_qos = session.will_qos
                payload.will_message = session.will_message
                payload.will_topic = session.will_topic
            else:
                vh.will_flag = False

            header = MQTTFixedHeader(PacketType.CONNECT, 0x00)
            packet = ConnectPacket(header, vh, payload)
            return packet

        packet = build_connect_packet(self.session)
        yield from self.outgoing_queue.put(packet)
        connack = yield from self.incoming_queues[PacketType.CONNACK].get()

        return connack.variable_header.return_code

    @asyncio.coroutine
    def mqtt_disconnect(self):
        disconnect_packet = DisconnectPacket()
        yield from self.outgoing_queue.put(disconnect_packet)
        self._ping_task.cancel()

    @asyncio.coroutine
    def mqtt_ping(self):
        self.logger.debug("Pinging ...")
        ping_packet = PingReqPacket()
        yield from self.outgoing_queue.put(ping_packet)
        yield from self.incoming_queues[PacketType.PINGRESP].get()



from enum import Enum

class SessionState(Enum):
    NEW = 0
    CONNECTED = 1
    DISCONNECTED = 2

class Session:
    def __init__(self):
        self.state = SessionState.NEW
        self.reader = None
        self.writer = None
        self.remote_address = None
        self.remote_port = None
        self.local_address = None
        self.local_port = None
        self.client_id = None
        self.clean_session = None
        self.will_flag = False
        self.will_message = None
        self.will_qos = None
        self.will_retain = None
        self.will_topic = None
        self.keep_alive = None
        self.username = None
        self.password = None
        self.scheme = None
        self._packet_id = 0

    @property
    def next_packet_id(self):
        self._packet_id += 1
        return self._packet_id



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
from hbmqtt.mqtt.packet import MQTTPacket, MQTTFixedHeader, PacketType, MQTTVariableHeader, MQTTPayload
from hbmqtt.errors import HBMQTTException, MQTTException
from hbmqtt.codecs import *


class PublishFixedHeader(MQTTFixedHeader):
    DUP_FLAG = 0x08
    RETAIN_FLAG = 0x01
    QOS_FLAG = 0x06

    def set_flags(self, dup_flag=False, qos=0, retain_flag=False):
        self.dup_flag = dup_flag
        self.retain_flag = retain_flag
        self.qos = qos

    def _set_flag(self, val, mask):
        if val:
            self.flags |= mask
        else:
            self.flags &= ~mask

    def _get_flag(self, mask):
        if self.flags & mask:
            return True
        else:
            return False

    @property
    def dup_flag(self) -> bool:
        return self._get_flag(self.DUP_FLAG)

    @dup_flag.setter
    def dup_flag(self, val: bool):
        self._set_flag(val, self.DUP_FLAG)

    @property
    def retain_flag(self) -> bool:
        return self._get_flag(self.RETAIN_FLAG)

    @retain_flag.setter
    def retain_flag(self, val: bool):
        self._set_flag(val, self.RETAIN_FLAG)

    @property
    def qos(self):
        return (self.flags & self.QOS_FLAG) >> 1

    @qos.setter
    def qos(self, val: int):
        self.flags &= (0x00 << 1)
        self.flags |= (val << 1)


class PublishVariableHeader(MQTTVariableHeader):
    def __init__(self, topic_name: str, packet_id: int=None):
        super().__init__()
        if '*' in topic_name:
            raise MQTTException("[MQTT-3.3.2-2] Topic name in the PUBLISH Packet MUST NOT contain wildcard characters.")
        self.topic_name = topic_name
        self.packet_id = packet_id

    def __repr__(self):
        return type(self).__name__ + '(topic={0}, packet_id={1})'.format(self.topic_name, self.packet_id)

    def to_bytes(self):
        out = b''
        out += encode_string(self.topic_name)
        if self.packet_id is not None:
            out += int_to_bytes(self.packet_id, 2)
        return out

    @classmethod
    def from_stream(cls, reader: asyncio.StreamReader, fixed_header: PublishFixedHeader):
        topic_name = yield from decode_string(reader)
        if fixed_header.qos:
            packet_id = yield from decode_packet_id(reader)
        else:
            packet_id = None
        return cls(topic_name, packet_id)


class PublishPayload(MQTTPayload):
    def __init__(self, data: bytes=None):
        super().__init__()
        self.data = data

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        return self.data

    @classmethod
    def from_stream(cls, reader: asyncio.StreamReader, fixed_header: MQTTFixedHeader,
                    variable_header: MQTTVariableHeader):
        data = yield from reader.read()
        return cls(data)


class PublishPacket(MQTTPacket):
    FIXED_HEADER = PublishFixedHeader
    VARIABLE_HEADER = PublishVariableHeader
    PAYLOAD = PublishPayload

    def __init__(self, fixed: PublishFixedHeader=None, variable_header: PublishVariableHeader=None, payload=None):
        if fixed is None:
            header = PublishFixedHeader(PacketType.PUBLISH, 0x00)
        else:
            if fixed.packet_type is not PacketType.PUBLISH:
                raise HBMQTTException("Invalid fixed packet type %s for PublishPacket init" % fixed.packet_type)
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, topic_name: str, message:bytes, packet_id: int, dup_flag, qos, retain):
        v_header = PublishVariableHeader(topic_name, packet_id)
        payload = PublishPayload(message)
        packet = PublishPacket(variable_header=v_header, payload=payload)
        packet.fixed_header.dup_flag = dup_flag
        packet.fixed_header.retain_flag = retain
        packet.fixed_header.qos = qos
        return packet



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import asyncio
from asyncio import IncompleteReadError
from math import ceil

from hbmqtt.errors import NoDataException


def bytes_to_hex_str(data):
    """
    converts a sequence of bytes into its displayable hex representation, ie: 0x??????
    :param data: byte sequence
    :return: Hexadecimal displayable representation
    """
    return '0x' + ''.join(format(b, '02x') for b in data)

def bytes_to_int(data):
    """
    convert a sequence of bytes to an integer using big endian byte ordering
    :param data: byte sequence
    :return: integer value
    """
    if isinstance(data, int):
        return data
    else:
        return int.from_bytes(data, byteorder='big')

def int_to_bytes(int_value: int, length=-1) -> bytes:
    """
    convert an integer to a sequence of bytes using big endian byte ordering
    :param int_value: integer value to convert
    :param length: (optional) byte length
    :return: byte sequence
    """
    if length == -1:
        length = ceil(int_value.bit_length()//8)
        if length == 0:
            length = 1
    return int_value.to_bytes(length, byteorder='big')


@asyncio.coroutine
def read_or_raise(reader, n=-1):
    """
    Read a given byte number from Stream. NoDataException is raised if read gives no data
    :param reader: Stream reader
    :param n: number of bytes to read
    :return: bytes read
    """
    try:
        data = yield from reader.readexactly(n)
    except IncompleteReadError:
        raise NoDataException("Incomplete read")
    if not data:
        raise NoDataException("No more data")
    return data

@asyncio.coroutine
def decode_string(reader) -> bytes:
    """
    Read a string from a reader and decode it according to MQTT string specification
    :param reader: Stream reader
    :return: UTF-8 string read from stream
    """
    length_bytes = yield from read_or_raise(reader, 2)
    str_length = bytes_to_int(length_bytes)
    byte_str = yield from read_or_raise(reader, str_length)
    return byte_str.decode(encoding='utf-8')

def encode_string(string: str) -> bytes:
    data = string.encode(encoding='utf-8')
    data_length = len(data)
    return int_to_bytes(data_length, 2) + data

@asyncio.coroutine
def decode_packet_id(reader) -> int:
    """
    Read a packet ID as 2-bytes int from stream according to MQTT specification (2.3.1)
    :param reader: Stream reader
    :return: Packet ID
    """
    packet_id_bytes = yield from read_or_raise(reader, 2)
    return bytes_to_int(packet_id_bytes)



# # Copyright (c) 2015 Nicolas JOUANIN
# #
# # See the file license.txt for copying permission.
# import asyncio
# import threading
# import logging
#
# from transitions import Machine, MachineError
#
# from hbmqtt.errors import BrokerException
# from hbmqtt.codecs.header import MQTTHeaderCodec
# from hbmqtt.codecs.connect import ConnectMessage
# from hbmqtt.messages.packet import PacketType, MQTTMessage, MQTTHeader
# from hbmqtt.errors import MQTTException
# from hbmqtt.session import Session
# from hbmqtt.broker.handlers import ConnectHandler
#
#
# class Broker:
#     states = ['new', 'starting', 'started', 'not_started', 'stopping', 'stopped', 'not_stopped', 'stopped']
#
#     def __init__(self, host='127.0.0.1', port=1883, loop = None):
#         self.logger = logging.getLogger(__name__)
#         self._init_states()
#         self.host = host
#         self.port = port
#         if loop is not None:
#             self._loop = loop
#         else:
#             self._loop = asyncio.get_event_loop()
#         self._server = None
#         self._loop_thread = None
#         self._message_handlers = None
#         self._codecs = None
#         self._sessions = dict()
#
#     def _init_states(self):
#         self.machine = Machine(states=Broker.states, initial='new')
#         self.machine.add_transition(trigger='start', source='new', dest='starting')
#         self.machine.add_transition(trigger='starting_fail', source='starting', dest='not_started')
#         self.machine.add_transition(trigger='starting_success', source='starting', dest='started')
#         self.machine.add_transition(trigger='shutdown', source='started', dest='stopping')
#         self.machine.add_transition(trigger='stopping_success', source='stopping', dest='stopped')
#         self.machine.add_transition(trigger='stopping_failure', source='stopping', dest='not_stopped')
#         self.machine.add_transition(trigger='start', source='stopped', dest='starting')
#
#
#     def _init_handlers(self):
#         self._message_handlers = {
#             PacketType.CONNECT, ConnectHandler(self),
#         }
#
#     def _init_codecs(self):
#         """
#         Init dict of MQTT message encoders/decoders
#         :return:
#         """
#         self._codecs = {
#             PacketType.CONNECT, ConnectMessage
#         }
#
#     def start(self):
#         try:
#             self.machine.start()
#             self.logger.debug("Broker starting")
#         except MachineError as me:
#             self.logger.debug("Invalid method call at this moment: %s" % me)
#             raise BrokerException("Broker instance can't be started: %s" % me)
#
#         try:
#             self._init_handlers()
#             self._init_codecs()
#             self._loop_thread = threading.Thread(target=self._run_server_loop, args=(self._loop,))
#             self._loop_thread.setDaemon(True)
#             self._loop_thread.start()
#         except Exception as e:
#             self.logger.error("Broker startup failed: %s" % e)
#             self.machine.starting_fail()
#             raise BrokerException("Broker instance can't be started: %s" % e)
#
#     def shutdown(self):
#         try:
#             self.machine.shutdown()
#         except MachineError as me:
#             self.logger.debug("Invalid method call at this moment: %s" % me)
#             raise BrokerException("Broker instance can't be stopped: %s" % me)
#         self._loop.call_soon_threadsafe(self._loop.stop)
#         self._loop_thread.join()
#         self._server.close()
#         self._loop.run_until_complete(self._server.wait_closed())
#         self._loop.close()
#         self._server = None
#         self._message_handlers = None
#         self._codecs = None
#         self.machine.stopping_success()
#
#     def _run_server_loop(self, loop):
#         asyncio.set_event_loop(loop)
#         coro = asyncio.start_server(self.client_connected, self.host, self.port)
#         self._server = loop.run_until_complete(coro)
#         self.logger.debug("Broker listening %s:%s" % (self.host, self.port))
#         self.machine.starting_success()
#         self.logger.debug("Broker started, ready to serve")
#         loop.run_forever()
#
#     def discard_session(self, client_id):
#         if client_id in self._sessions:
#             del self._sessions[client_id]
#         else:
#             self.logger.warn("Asked to discard an unknown client ID session")
#
#     def create_session(self, remote_address, remote_port, client_id, clean_session):
#         session = Session(remote_address, remote_port, client_id, clean_session)
#         if client_id in self._sessions:
#             raise BrokerException("Session already exists for client ID: %s", client_id)
#         self.save_session(session)
#         return session
#
#     def get_session(self, client_id):
#         if client_id not in self._sessions:
#             raise BrokerException("Unknown session for client ID: %s", client_id)
#         else:
#             return self._sessions[client_id]
#
#     def resume_session(self, session: Session):
#         # TBD
#         pass
#
#     def save_session(self, session: Session):
#         self._sessions[session.client_id] = session
#
#     @asyncio.coroutine
#     def _handle_message(self, session: Session, message: MQTTMessage) -> MQTTMessage:
#         handler = self._message_handlers[message.mqtt_header.message_type]
#         response = yield from handler.handle(session, message)
#         return response
#
#     @asyncio.coroutine
#     def decode_message(self, header: MQTTHeader, reader):
#         decoder = self._codecs[header.message_type]
#         message = yield from decoder.decode(header, reader)
#         return message
#
#     @asyncio.coroutine
#     def _encode_message(self, message: MQTTMessage) -> bytes:
#         encoder = self._codecs[message.mqtt_header.message_type]
#         out = yield from encoder.encode(message)
#         message.mqtt_header.remaining_length = len(out)
#         encoded = MQTTHeaderCodec.encode(message.mqtt_header) + out
#         return encoded
#
#     @asyncio.coroutine
#     def run_for_messages(self, session):
#         header = yield from MQTTHeaderCodec.decode(session.reader)
#         handler = self._message_handlers[header.message_type]
#         message = yield from handler.
#
#     @asyncio.coroutine
#     def client_connected(self, reader, writer):
#         # Init new session
#         (remote_address, remote_port) = writer.get_extra_info('peername')
#         session = Session(reader, writer, remote_address, remote_port)
#
#         while True:
#             yield from self.run_for_messages(session)
#             except MQTTException:
#                 # End connection
#                 break
#
#
#         # Handle connection
#         try:
#             header = yield from MQTTHeaderCodec.decode(reader)
#             if header.message_type != PacketType.CONNECT:
#                 raise MQTTException("[MQTT-3.1.0-1] First Packet sent from the Client MUST be a CONNECT Packet")
#             request = yield from self._decode_message(header.message_type, reader)
#
#             (remote_address, remote_port) = writer.get_extra_info('peername')
#             session = Session(reader, writer, remote_address, remote_port)
#
#             response = self._handle_message(session, request)
#             encoded_response = yield from self._encode_message(response)
#             writer.write(encoded_response)
#             yield from writer.drain()
#         except MQTTException:
#             return
#
#         # Enter message loop
#         while True:
#             try:
#                 header = yield from MQTTHeaderCodec.decode(reader)
#                 request = yield from self._decode_message(header.message_type, reader)
#                 response = self._handle_message(request)
#                 encoded_response = yield from self._encode_message(response)
#                 writer.write(encoded_response)
#                 yield from writer.drain()
#             except MQTTException:
#                 # End connection
#                 break



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest

from hbmqtt.mqtt.subscribe import SubscribePacket, SubscribePayload
from hbmqtt.mqtt.packet import PacketIdVariableHeader
from hbmqtt.codecs import *

class SubscribePacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_stream(self):
        data = b'\x80\x0e\x00\x0a\x00\x03a/b\x01\x00\x03c/d\x02'
        stream = asyncio.StreamReader(loop=self.loop)
        stream.feed_data(data)
        stream.feed_eof()
        message = self.loop.run_until_complete(SubscribePacket.from_stream(stream))
        self.assertEqual(message.payload.topics[0]['filter'], 'a/b')
        self.assertEqual(message.payload.topics[0]['qos'], 0x01)
        self.assertEqual(message.payload.topics[1]['filter'], 'c/d')
        self.assertEqual(message.payload.topics[1]['qos'], 0x02)

    def test_to_stream(self):
        variable_header = PacketIdVariableHeader(10)
        payload = SubscribePayload(
            [
                {'filter': 'a/b', 'qos': 0x01},
                {'filter': 'c/d', 'qos': 0x02}
            ])
        publish = SubscribePacket(variable_header=variable_header, payload=payload)
        out = publish.to_bytes()
        self.assertEqual(out, b'\x82\x0e\x00\x0a\x00\x03a/b\x01\x00\x03c/d\x02')


# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
from hbmqtt.errors import HBMQTTException
from hbmqtt.mqtt.packet import MQTTFixedHeader, MQTTPacket, PacketType
from hbmqtt.mqtt.connect import ConnectPacket
from hbmqtt.mqtt.connack import ConnackPacket
from hbmqtt.mqtt.disconnect import DisconnectPacket
from hbmqtt.mqtt.pingreq import PingReqPacket
from hbmqtt.mqtt.pingresp import PingRespPacket
from hbmqtt.mqtt.publish import PublishPacket
from hbmqtt.mqtt.puback import PubackPacket
from hbmqtt.mqtt.pubrec import PubrecPacket
from hbmqtt.mqtt.pubrel import PubrelPacket
from hbmqtt.mqtt.pubcomp import PubcompPacket
from hbmqtt.mqtt.subscribe import SubscribePacket
from hbmqtt.mqtt.suback import SubackPacket
from hbmqtt.mqtt.unsubscribe import UnsubscribePacket
from hbmqtt.mqtt.unsuback import UnsubackPacket

packet_dict = {
    PacketType.CONNECT: ConnectPacket,
    PacketType.CONNACK: ConnackPacket,
    PacketType.PUBLISH: PublishPacket,
    PacketType.PUBACK: PubackPacket,
    PacketType.PUBREC: PubrecPacket,
    PacketType.PUBREL: PubrelPacket,
    PacketType.PUBCOMP: PubcompPacket,
    PacketType.SUBSCRIBE: SubscribePacket,
    PacketType.SUBACK: SubackPacket,
    PacketType.UNSUBSCRIBE: UnsubscribePacket,
    PacketType.UNSUBACK: UnsubackPacket,
    PacketType.PINGREQ: PingReqPacket,
    PacketType.PINGRESP: PingRespPacket,
    PacketType.DISCONNECT: DisconnectPacket
}

def packet_class(fixed_header: MQTTFixedHeader):
    try:
        cls = packet_dict[fixed_header.packet_type]
        return cls
    except KeyError:
        raise HBMQTTException("Unexpected packet Type '%s'" % fixed_header.packet_type)


# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest

from hbmqtt.mqtt.publish import PublishPacket, PublishVariableHeader, PublishPayload
from hbmqtt.codecs import *

class PublishPacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_stream(self):
        data = b'\x3f\x09\x00\x05topic\x00\x0a0123456789'
        stream = asyncio.StreamReader(loop=self.loop)
        stream.feed_data(data)
        stream.feed_eof()
        message = self.loop.run_until_complete(PublishPacket.from_stream(stream))
        self.assertEqual(message.variable_header.topic_name, 'topic')
        self.assertEqual(message.variable_header.packet_id, 10)
        self.assertEqual(message.fixed_header.qos, 0x03)
        self.assertTrue(message.fixed_header.dup_flag)
        self.assertTrue(message.fixed_header.retain_flag)
        self.assertTrue(message.payload.data, b'0123456789')

    def test_to_stream(self):
        variable_header = PublishVariableHeader('topic', 10)
        payload = PublishPayload(b'0123456789')
        publish = PublishPacket(variable_header=variable_header, payload=payload)
        out = publish.to_bytes()
        self.assertEqual(out, b'\x30\x13\x00\x05topic\x00\x0a0123456789')


# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest

from hbmqtt.mqtt.unsubscribe import UnsubscribePacket, UnubscribePayload
from hbmqtt.mqtt.packet import PacketIdVariableHeader
from hbmqtt.codecs import *

class UnsubscribePacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_stream(self):
        data = b'\xa2\x0c\x00\n\x00\x03a/b\x00\x03c/d'
        stream = asyncio.StreamReader(loop=self.loop)
        stream.feed_data(data)
        stream.feed_eof()
        message = self.loop.run_until_complete(UnsubscribePacket.from_stream(stream))
        self.assertEqual(message.payload.topics[0], 'a/b')
        self.assertEqual(message.payload.topics[1], 'c/d')

    def test_to_stream(self):
        variable_header = PacketIdVariableHeader(10)
        payload = UnubscribePayload(['a/b', 'c/d'])
        publish = UnsubscribePacket(variable_header=variable_header, payload=payload)
        out = publish.to_bytes()
        self.assertEqual(out, b'\xa2\x0c\x00\n\x00\x03a/b\x00\x03c/d')


# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
# from hbmqtt.errors import BrokerException
#
# class HandlerException(BrokerException):
#     """
#     Exceptions thrown by encode/decode process
#     """
#     pass
#



__author__ = 'nico'


def not_in_dict_or_none(dict, key):
    """
    Check if a key exists in a map and if it's not None
    :param dict: map to look for key
    :param key: key to find
    :return: true if key is in dict and not None
    """
    if key not in dict or dict[key] is None:
        return True
    else:
        return False


# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest

from hbmqtt.mqtt.suback import SubackPacket, SubackPayload
from hbmqtt.mqtt.packet import PacketIdVariableHeader
from hbmqtt.codecs import *

class SubackPacketTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_stream(self):
        data = b'\x90\x06\x00\x0a\x00\x01\x02\x80'
        stream = asyncio.StreamReader(loop=self.loop)
        stream.feed_data(data)
        stream.feed_eof()
        message = self.loop.run_until_complete(SubackPacket.from_stream(stream))
        self.assertEqual(message.payload.return_codes[0], SubackPayload.RETURN_CODE_00)
        self.assertEqual(message.payload.return_codes[1], SubackPayload.RETURN_CODE_01)
        self.assertEqual(message.payload.return_codes[2], SubackPayload.RETURN_CODE_02)
        self.assertEqual(message.payload.return_codes[3], SubackPayload.RETURN_CODE_80)

    def test_to_stream(self):
        variable_header = PacketIdVariableHeader(10)
        payload = SubackPayload(
            [SubackPayload.RETURN_CODE_00, SubackPayload.RETURN_CODE_01, SubackPayload.RETURN_CODE_02, SubackPayload.RETURN_CODE_80
            ])
        publish = SubackPacket(variable_header=variable_header, payload=payload)
        out = publish.to_bytes()
        self.assertEqual(out, b'\x90\x06\x00\x0a\x00\x01\x02\x80')